@@ -1,32 +1,39 @@
 "Fief client for Python."
+
 from fief_client.client import (
     Fief,
+    FiefAccessTokenACRTooLow,
     FiefAccessTokenExpired,
     FiefAccessTokenInfo,
     FiefAccessTokenInvalid,
     FiefAccessTokenMissingPermission,
     FiefAccessTokenMissingScope,
+    FiefACR,
     FiefAsync,
     FiefError,
     FiefIdTokenInvalid,
+    FiefRequestError,
     FiefTokenResponse,
     FiefUserInfo,
 )
 
-__version__ = "0.15.0"
+__version__ = "0.20.0"
 
 __all__ = [
     "Fief",
+    "FiefACR",
     "FiefAsync",
     "FiefTokenResponse",
     "FiefAccessTokenInfo",
     "FiefUserInfo",
     "FiefError",
+    "FiefAccessTokenACRTooLow",
     "FiefAccessTokenExpired",
     "FiefAccessTokenMissingPermission",
     "FiefAccessTokenMissingScope",
     "FiefAccessTokenInvalid",
     "FiefIdTokenInvalid",
+    "FiefRequestError",
     "crypto",
     "pkce",
     "integrations",
