@@ -1,15 +1,10 @@
-import sys
-import uuid
-
-if sys.version_info < (3, 8):
-    from typing_extensions import TypedDict  # pragma: no cover
-else:
-    from typing import TypedDict  # pragma: no cover
-
 import contextlib
 import json
-from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
-from urllib.parse import urlencode
+import uuid
+from collections.abc import Mapping
+from enum import Enum
+from typing import Any, Optional, TypedDict, Union
+from urllib.parse import urlencode, urlsplit, urlunsplit
 
 import httpx
 from httpx._types import CertTypes, VerifyTypes
@@ -20,6 +15,43 @@ from fief_client.crypto import is_valid_hash
 HTTPXClient = Union[httpx.Client, httpx.AsyncClient]
 
 
+class FiefACR(str, Enum):
+    """
+    List of defined Authentication Context Class Reference.
+    """
+
+    LEVEL_ZERO = "0"
+    """Level 0. No authentication was performed, a previous session was used."""
+    LEVEL_ONE = "1"
+    """Level 1. Password authentication was performed."""
+
+    def __lt__(self, other: object) -> bool:
+        return self._compare(other, True, True)
+
+    def __le__(self, other: object) -> bool:
+        return self._compare(other, False, True)
+
+    def __gt__(self, other: object) -> bool:
+        return self._compare(other, True, False)
+
+    def __ge__(self, other: object) -> bool:
+        return self._compare(other, False, False)
+
+    def _compare(self, other: object, strict: bool, asc: bool) -> bool:
+        if not isinstance(other, FiefACR):
+            return NotImplemented  # pragma: no cover
+
+        if self == other:
+            return not strict
+
+        for elem in FiefACR:
+            if self == elem:
+                return asc
+            elif other == elem:
+                return not asc
+        raise RuntimeError()  # pragma: no cover
+
+
 class FiefTokenResponse(TypedDict):
     """
     Typed dictionary containing the tokens and related information returned by Fief after a successful authentication.
@@ -47,6 +79,7 @@ class FiefAccessTokenInfo(TypedDict):
     {
         "id": "aeeb8bfa-e8f4-4724-9427-c3d5af66190e",
         "scope": ["openid", "required_scope"],
+        "acr": "1",
         "permissions": ["castles:read", "castles:create", "castles:update", "castles:delete"],
         "access_token": "ACCESS_TOKEN",
     }
@@ -55,9 +88,11 @@ class FiefAccessTokenInfo(TypedDict):
 
     id: uuid.UUID
     """ID of the user."""
-    scope: List[str]
+    scope: list[str]
     """List of granted scopes for this access token."""
-    permissions: List[str]
+    acr: FiefACR
+    """Level of Authentication Context class Reference."""
+    permissions: list[str]
     """List of [granted permissions](https://docs.fief.dev/getting-started/access-control/) for this user."""
     access_token: str
     """Access token you can use to call the Fief API."""
@@ -94,7 +129,7 @@ class FiefUserInfo(TypedDict):
     """
     ID of the [tenant](https://docs.fief.dev/getting-started/tenants/) associated to the user.
     """
-    fields: Dict[str, Any]
+    fields: dict[str, Any]
     """
     [User fields](https://docs.fief.dev/getting-started/user-fields/) values for this user, indexed by their slug.
     """
@@ -104,6 +139,16 @@ class FiefError(Exception):
     """Base Fief client error."""
 
 
+class FiefRequestError(FiefError):
+    """The request to Fief server resulted in an error."""
+
+    def __init__(self, status_code: int, detail: str) -> None:
+        self.status_code = status_code
+        self.detail = detail
+        self.message = f"[{status_code}] - {detail}"
+        super().__init__(self.message)
+
+
 class FiefAccessTokenInvalid(FiefError):
     """The access token is invalid."""
 
@@ -116,6 +161,10 @@ class FiefAccessTokenMissingScope(FiefError):
     """The access token is missing a required scope."""
 
 
+class FiefAccessTokenACRTooLow(FiefError):
+    """The access token doesn't meet the minimum ACR level."""
+
+
 class FiefAccessTokenMissingPermission(FiefError):
     """The access token is missing a required permission."""
 
@@ -133,12 +182,18 @@ class BaseFief:
     """Base URL of your Fief tenant."""
     client_id: str
     """ID of your Fief client."""
-    client_secret: str
-    """Secret of your Fief client."""
+    client_secret: Optional[str] = None
+    """
+    Secret of your Fief client.
+
+    If you're implementing a desktop app, it's not recommended to use it,
+    since it can be easily found by the end-user in the source code.
+    The recommended way is to use a [Public client](https://docs.fief.dev/getting-started/clients/#public-clients).
+    """
     encryption_key: Optional[jwk.JWK] = None
     """"""
 
-    _openid_configuration: Optional[Dict[str, Any]] = None
+    _openid_configuration: Optional[dict[str, Any]] = None
     _jwks: Optional[jwk.JWKSet] = None
 
     _verify: VerifyTypes
@@ -148,7 +203,7 @@ class BaseFief:
         self,
         base_url: str,
         client_id: str,
-        client_secret: str,
+        client_secret: Optional[str] = None,
         *,
         encryption_key: Optional[str] = None,
         host: Optional[str] = None,
@@ -161,6 +216,9 @@ class BaseFief:
         :param base_url: Base URL of your Fief tenant.
         :param client_id: ID of your Fief client.
         :param client_secret: Secret of your Fief client.
+        If you're implementing a desktop app, it's not recommended to use it,
+        since it can be easily found by the end-user in the source code.
+        The recommended way is to use a [Public client](https://docs.fief.dev/getting-started/clients/#public-clients).
         :param encryption_key: Encryption key of your Fief client.
         Necessary only if [ID Token encryption](https://docs.fief.dev/going-further/id-token-encryption/) is enabled.
         :param verify: Corresponds to the [verify parameter of HTTPX](https://www.python-httpx.org/advanced/#changing-the-verification-defaults).
@@ -178,17 +236,37 @@ class BaseFief:
         self.cert = cert
 
     def _get_endpoint_url(
-        self, openid_configuration: Dict[str, Any], field: str
+        self,
+        openid_configuration: dict[str, Any],
+        field: str,
+        *,
+        absolute: bool = False,
     ) -> str:
+        """
+        Return the specified endpoint from OpenID configuration.
+
+        If `absolute` is `False`, we only retain the path,
+        as the host might is not always relevant in our context.
+
+        Typically, we might be in a Docker environment where the client app has to make
+        request to the Fief server through Docker networking. Therefore, we do not
+        want the client to use the absolute URL generated by OpenID Configuration, but
+        rather stick to the host specified on the client configuration.
+        """
+        if not absolute:
+            (scheme, netloc, *components) = urlsplit(self.base_url)
+            host = self.host if self.host is not None else netloc
+            host_base_url = urlunsplit((scheme, host, *components))
+            return openid_configuration[field].split(host_base_url)[1]
         return openid_configuration[field]
 
     def _auth_url(
         self,
-        openid_configuration: Dict[str, Any],
+        openid_configuration: dict[str, Any],
         redirect_uri: str,
         *,
         state: Optional[str] = None,
-        scope: Optional[List[str]] = None,
+        scope: Optional[list[str]] = None,
         code_challenge: Optional[str] = None,
         code_challenge_method: Optional[str] = None,
         lang: Optional[str] = None,
@@ -217,7 +295,7 @@ class BaseFief:
             params = {**params, **extras_params}
 
         authorization_endpoint = self._get_endpoint_url(
-            openid_configuration, "authorization_endpoint"
+            openid_configuration, "authorization_endpoint", absolute=True
         )
         return f"{authorization_endpoint}?{urlencode(params)}"
 
@@ -226,8 +304,9 @@ class BaseFief:
         access_token: str,
         jwks: jwk.JWKSet,
         *,
-        required_scope: Optional[List[str]] = None,
-        required_permissions: Optional[List[str]] = None,
+        required_scope: Optional[list[str]] = None,
+        required_acr: Optional[FiefACR] = None,
+        required_permissions: Optional[list[str]] = None,
     ) -> FiefAccessTokenInfo:
         try:
             decoded_token = jwt.JWT(jwt=access_token, algs=["RS256"], key=jwks)
@@ -238,7 +317,16 @@ class BaseFief:
                     if scope not in access_token_scope:
                         raise FiefAccessTokenMissingScope()
 
-            permissions: List[str] = claims["permissions"]
+            try:
+                acr = FiefACR(claims["acr"])
+            except ValueError as e:
+                raise FiefAccessTokenInvalid() from e
+
+            if required_acr is not None:
+                if acr < required_acr:
+                    raise FiefAccessTokenACRTooLow()
+
+            permissions: list[str] = claims["permissions"]
             if required_permissions is not None:
                 for required_permission in required_permissions:
                     if required_permission not in permissions:
@@ -247,15 +335,14 @@ class BaseFief:
             return {
                 "id": uuid.UUID(claims["sub"]),
                 "scope": access_token_scope,
+                "acr": acr,
                 "permissions": permissions,
                 "access_token": access_token,
             }
 
         except jwt.JWTExpired as e:
             raise FiefAccessTokenExpired() from e
-        except jwt.JWException as e:
-            raise FiefAccessTokenInvalid() from e
-        except KeyError as e:
+        except (jwt.JWException, KeyError, ValueError) as e:
             raise FiefAccessTokenInvalid() from e
 
     def _decode_id_token(
@@ -286,9 +373,10 @@ class BaseFief:
                 ):
                     raise FiefIdTokenInvalid()
 
-            return claims
         except (jwt.JWException, TypeError) as e:
             raise FiefIdTokenInvalid() from e
+        else:
+            return claims
 
     def _get_openid_configuration_request(self, client: HTTPXClient) -> httpx.Request:
         return client.build_request("GET", "/.well-known/openid-configuration")
@@ -302,20 +390,17 @@ class BaseFief:
         redirect_uri: str,
         code_verifier: Optional[str] = None,
     ) -> httpx.Request:
-        basic_auth = httpx.BasicAuth(self.client_id, self.client_secret)
-        return client.build_request(
-            "POST",
-            endpoint,
-            headers={
-                "Authorization": basic_auth._auth_header,
-            },
-            data={
-                "grant_type": "authorization_code",
-                "code": code,
-                "redirect_uri": redirect_uri,
-                "code_verifier": code_verifier,
-            },
-        )
+        data = {
+            "client_id": self.client_id,
+            "grant_type": "authorization_code",
+            "code": code,
+            "redirect_uri": redirect_uri,
+        }
+        if code_verifier is not None:
+            data["code_verifier"] = code_verifier
+        if self.client_secret is not None:
+            data["client_secret"] = self.client_secret
+        return client.build_request("POST", endpoint, data=data)
 
     def _get_auth_refresh_token_request(
         self,
@@ -323,24 +408,19 @@ class BaseFief:
         *,
         endpoint: str,
         refresh_token: str,
-        scope: Optional[List[str]] = None,
+        scope: Optional[list[str]] = None,
     ) -> httpx.Request:
         data = {
+            "client_id": self.client_id,
             "grant_type": "refresh_token",
             "refresh_token": refresh_token,
         }
+        if self.client_secret is not None:
+            data["client_secret"] = self.client_secret
         if scope is not None:
             data["scope"] = " ".join(scope)
 
-        basic_auth = httpx.BasicAuth(self.client_id, self.client_secret)
-        return client.build_request(
-            "POST",
-            endpoint,
-            headers={
-                "Authorization": basic_auth._auth_header,
-            },
-            data=data,
-        )
+        return client.build_request("POST", endpoint, data=data)
 
     def _get_userinfo_request(
         self, client: HTTPXClient, *, endpoint: str, access_token: str
@@ -355,7 +435,7 @@ class BaseFief:
         *,
         endpoint: str,
         access_token: str,
-        data: Dict[str, Any],
+        data: dict[str, Any],
     ) -> httpx.Request:
         return client.build_request(
             "PATCH",
@@ -364,6 +444,55 @@ class BaseFief:
             json=data,
         )
 
+    def _get_change_password_request(
+        self,
+        client: HTTPXClient,
+        *,
+        endpoint: str,
+        access_token: str,
+        new_password: str,
+    ) -> httpx.Request:
+        return client.build_request(
+            "PATCH",
+            endpoint,
+            headers={"Authorization": f"Bearer {access_token}"},
+            json={"password": new_password},
+        )
+
+    def _get_email_change_request(
+        self,
+        client: HTTPXClient,
+        *,
+        endpoint: str,
+        access_token: str,
+        email: str,
+    ) -> httpx.Request:
+        return client.build_request(
+            "PATCH",
+            endpoint,
+            headers={"Authorization": f"Bearer {access_token}"},
+            json={"email": email},
+        )
+
+    def _get_email_verify_request(
+        self,
+        client: HTTPXClient,
+        *,
+        endpoint: str,
+        access_token: str,
+        code: str,
+    ) -> httpx.Request:
+        return client.build_request(
+            "POST",
+            endpoint,
+            headers={"Authorization": f"Bearer {access_token}"},
+            json={"code": code},
+        )
+
+    def _handle_request_error(self, response: httpx.Response):
+        if response.is_error:
+            raise FiefRequestError(response.status_code, response.text)
+
 
 class Fief(BaseFief):
     """Sync Fief authentication client."""
@@ -372,7 +501,7 @@ class Fief(BaseFief):
         self,
         base_url: str,
         client_id: str,
-        client_secret: str,
+        client_secret: Optional[str] = None,
         *,
         encryption_key: Optional[str] = None,
         host: Optional[str] = None,
@@ -394,7 +523,7 @@ class Fief(BaseFief):
         redirect_uri: str,
         *,
         state: Optional[str] = None,
-        scope: Optional[List[str]] = None,
+        scope: Optional[list[str]] = None,
         code_challenge: Optional[str] = None,
         code_challenge_method: Optional[str] = None,
         lang: Optional[str] = None,
@@ -412,7 +541,7 @@ class Fief(BaseFief):
         :param lang: Optional parameter to set the user locale.
         Should be a valid [RFC 3066](https://www.rfc-editor.org/rfc/rfc3066) language identifier, like `fr` or `pt-PT`.
         If not provided, the user locale is determined by their browser settings.
-        :param extras_params: Optional dictionary containing specific parameters.
+        :param extras_params: Optional dictionary containing [specific parameters](https://docs.fief.dev/going-further/authorize-url/).
 
         **Example:**
 
@@ -434,7 +563,7 @@ class Fief(BaseFief):
 
     def auth_callback(
         self, code: str, redirect_uri: str, *, code_verifier: Optional[str] = None
-    ) -> Tuple[FiefTokenResponse, FiefUserInfo]:
+    ) -> tuple[FiefTokenResponse, FiefUserInfo]:
         """
         Return a `FiefTokenResponse` and `FiefUserInfo` in exchange of an authorization code.
 
@@ -462,8 +591,8 @@ class Fief(BaseFief):
         return token_response, userinfo
 
     def auth_refresh_token(
-        self, refresh_token: str, *, scope: Optional[List[str]] = None
-    ) -> Tuple[FiefTokenResponse, FiefUserInfo]:
+        self, refresh_token: str, *, scope: Optional[list[str]] = None
+    ) -> tuple[FiefTokenResponse, FiefUserInfo]:
         """
         Return fresh `FiefTokenResponse` and `FiefUserInfo` in exchange of a refresh token
 
@@ -490,7 +619,7 @@ class Fief(BaseFief):
             )
             response = client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
 
             token_response = response.json()
         jwks = self._get_jwks()
@@ -505,8 +634,9 @@ class Fief(BaseFief):
         self,
         access_token: str,
         *,
-        required_scope: Optional[List[str]] = None,
-        required_permissions: Optional[List[str]] = None,
+        required_scope: Optional[list[str]] = None,
+        required_acr: Optional[FiefACR] = None,
+        required_permissions: Optional[list[str]] = None,
     ) -> FiefAccessTokenInfo:
         """
         Check if an access token is valid and optionally that it has a required list of scopes,
@@ -515,6 +645,8 @@ class Fief(BaseFief):
 
         :param access_token: The access token to validate.
         :param required_scope: Optional list of scopes to check for.
+        :param required_acr: Optional minimum ACR level required.
+        Read more: https://docs.fief.dev/going-further/acr/
         :param required_permissions: Optional list of permissions to check for.
 
         **Example: Validate access token with required scopes**
@@ -532,6 +664,21 @@ class Fief(BaseFief):
         print(access_token_info)
         ```
 
+        **Example: Validate access token with minimum ACR level**
+
+        ```py
+        try:
+            access_token_info = fief.validate_access_token("ACCESS_TOKEN", required_acr=FiefACR.LEVEL_ONE)
+        except FiefAccessTokenInvalid:
+            print("Invalid access token")
+        except FiefAccessTokenExpired:
+            print("Expired access token")
+        except FiefAccessTokenACRTooLow:
+            print("ACR too low")
+
+        print(access_token_info)
+        ```
+
         **Example: Validate access token with required permissions**
 
         ```py
@@ -552,6 +699,7 @@ class Fief(BaseFief):
             access_token,
             jwks,
             required_scope=required_scope,
+            required_acr=required_acr,
             required_permissions=required_permissions,
         )
 
@@ -576,29 +724,17 @@ class Fief(BaseFief):
             )
             response = client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
 
             return response.json()
 
-    def update_profile(self, access_token: str, data: Dict[str, Any]) -> FiefUserInfo:
+    def update_profile(self, access_token: str, data: dict[str, Any]) -> FiefUserInfo:
         """
         Update user information with the Fief API using a valid access token.
 
         :param access_token: A valid access token.
         :param data: A dictionary containing the data to update.
 
-        **Example: Update email address**
-
-        ```py
-        userinfo = fief.update_profile("ACCESS_TOKEN", { "email": "anne@nantes.city" })
-        ```
-
-        **Example: Update password**
-
-        ```py
-        userinfo = fief.update_profile("ACCESS_TOKEN", { "password": "hermine1" })
-        ```
-
         **Example: Update user field**
 
         To update [user field](https://docs.fief.dev/getting-started/user-fields/) values, you need to nest them into a `fields` dictionary, indexed by their slug.
@@ -618,7 +754,100 @@ class Fief(BaseFief):
             )
             response = client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
+
+            return response.json()
+
+    def change_password(self, access_token: str, new_password: str) -> FiefUserInfo:
+        """
+        Change the user password with the Fief API using a valid access token.
+
+        **An access token with an ACR of at least level 1 is required.**
+
+        :param access_token: A valid access token.
+        :param new_password: The new password.
+
+        **Example**
+
+        ```py
+        userinfo = fief.change_password("ACCESS_TOKEN", "herminetincture")
+        ```
+        """
+        change_password_profile_endpoint = f"{self.base_url}/api/password"
+
+        with self._get_httpx_client() as client:
+            request = self._get_change_password_request(
+                client,
+                endpoint=change_password_profile_endpoint,
+                access_token=access_token,
+                new_password=new_password,
+            )
+            response = client.send(request)
+
+            self._handle_request_error(response)
+
+            return response.json()
+
+    def email_change(self, access_token: str, email: str) -> FiefUserInfo:
+        """
+        Request an email change with the Fief API using a valid access token.
+
+        The user will receive a verification code on this new email address.
+        It shall be used with the method `email_verify` to complete the modification.
+
+        **An access token with an ACR of at least level 1 is required.**
+
+        :param access_token: A valid access token.
+        :param email: The new email address.
+
+        **Example**
+
+        ```py
+        userinfo = fief.email_change("ACCESS_TOKEN", "anne@nantes.city")
+        ```
+        """
+        email_change_endpoint = f"{self.base_url}/api/email/change"
+
+        with self._get_httpx_client() as client:
+            request = self._get_email_change_request(
+                client,
+                endpoint=email_change_endpoint,
+                access_token=access_token,
+                email=email,
+            )
+            response = client.send(request)
+
+            self._handle_request_error(response)
+
+            return response.json()
+
+    def email_verify(self, access_token: str, code: str) -> FiefUserInfo:
+        """
+        Verify the user email with the Fief API using a valid access token and verification code.
+
+        **An access token with an ACR of at least level 1 is required.**
+
+        :param access_token: A valid access token.
+        :param code: The verification code received by email.
+
+        **Example**
+
+        ```py
+        userinfo = fief.email_verify("ACCESS_TOKEN", "ABCDE")
+        ```
+        """
+        email_verify_endpoint = f"{self.base_url}/api/email/verify"
+
+        with self._get_httpx_client() as client:
+            request = self._get_email_verify_request(
+                client,
+                endpoint=email_verify_endpoint,
+                access_token=access_token,
+                code=code,
+            )
+            response = client.send(request)
+
+            self._handle_request_error(response)
 
             return response.json()
 
@@ -650,7 +879,7 @@ class Fief(BaseFief):
         ) as client:
             yield client
 
-    def _get_openid_configuration(self) -> Dict[str, Any]:
+    def _get_openid_configuration(self) -> dict[str, Any]:
         if self._openid_configuration is not None:
             return self._openid_configuration
 
@@ -687,7 +916,7 @@ class Fief(BaseFief):
             )
             response = client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
 
             return response.json()
 
@@ -699,7 +928,7 @@ class FiefAsync(BaseFief):
         self,
         base_url: str,
         client_id: str,
-        client_secret: str,
+        client_secret: Optional[str] = None,
         *,
         encryption_key: Optional[str] = None,
         host: Optional[str] = None,
@@ -721,7 +950,7 @@ class FiefAsync(BaseFief):
         redirect_uri: str,
         *,
         state: Optional[str] = None,
-        scope: Optional[List[str]] = None,
+        scope: Optional[list[str]] = None,
         code_challenge: Optional[str] = None,
         code_challenge_method: Optional[str] = None,
         lang: Optional[str] = None,
@@ -739,7 +968,7 @@ class FiefAsync(BaseFief):
         :param lang: Optional parameter to set the user locale.
         Should be a valid [RFC 3066](https://www.rfc-editor.org/rfc/rfc3066) language identifier, like `fr` or `pt-PT`.
         If not provided, the user locale is determined by their browser settings.
-        :param extras_params: Optional dictionary containing specific parameters.
+        :param extras_params: Optional dictionary containing [specific parameters](https://docs.fief.dev/going-further/authorize-url/).
 
         **Example:**
 
@@ -761,7 +990,7 @@ class FiefAsync(BaseFief):
 
     async def auth_callback(
         self, code: str, redirect_uri: str, *, code_verifier: Optional[str] = None
-    ) -> Tuple[FiefTokenResponse, FiefUserInfo]:
+    ) -> tuple[FiefTokenResponse, FiefUserInfo]:
         """
         Return a `FiefTokenResponse` and `FiefUserInfo` in exchange of an authorization code.
 
@@ -789,8 +1018,8 @@ class FiefAsync(BaseFief):
         return token_response, userinfo
 
     async def auth_refresh_token(
-        self, refresh_token: str, *, scope: Optional[List[str]] = None
-    ) -> Tuple[FiefTokenResponse, FiefUserInfo]:
+        self, refresh_token: str, *, scope: Optional[list[str]] = None
+    ) -> tuple[FiefTokenResponse, FiefUserInfo]:
         """
         Return fresh `FiefTokenResponse` and `FiefUserInfo` in exchange of a refresh token
 
@@ -817,7 +1046,7 @@ class FiefAsync(BaseFief):
             )
             response = await client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
 
             token_response = response.json()
 
@@ -833,8 +1062,9 @@ class FiefAsync(BaseFief):
         self,
         access_token: str,
         *,
-        required_scope: Optional[List[str]] = None,
-        required_permissions: Optional[List[str]] = None,
+        required_scope: Optional[list[str]] = None,
+        required_acr: Optional[FiefACR] = None,
+        required_permissions: Optional[list[str]] = None,
     ) -> FiefAccessTokenInfo:
         """
         Check if an access token is valid and optionally that it has a required list of scopes,
@@ -843,6 +1073,8 @@ class FiefAsync(BaseFief):
 
         :param access_token: The access token to validate.
         :param required_scope: Optional list of scopes to check for.
+        :param required_acr: Optional minimum ACR level required.
+        Read more: https://docs.fief.dev/going-further/acr/
         :param required_permissions: Optional list of permissions to check for.
 
         **Example: Validate access token with required scopes**
@@ -860,6 +1092,21 @@ class FiefAsync(BaseFief):
         print(access_token_info)
         ```
 
+        **Example: Validate access token with minimum ACR level**
+
+        ```py
+        try:
+            access_token_info = await fief.validate_access_token("ACCESS_TOKEN", required_acr=FiefACR.LEVEL_ONE)
+        except FiefAccessTokenInvalid:
+            print("Invalid access token")
+        except FiefAccessTokenExpired:
+            print("Expired access token")
+        except FiefAccessTokenACRTooLow:
+            print("ACR too low")
+
+        print(access_token_info)
+        ```
+
         **Example: Validate access token with required permissions**
 
         ```py
@@ -880,6 +1127,7 @@ class FiefAsync(BaseFief):
             access_token,
             jwks,
             required_scope=required_scope,
+            required_acr=required_acr,
             required_permissions=required_permissions,
         )
 
@@ -904,12 +1152,12 @@ class FiefAsync(BaseFief):
             )
             response = await client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
 
             return response.json()
 
     async def update_profile(
-        self, access_token: str, data: Dict[str, Any]
+        self, access_token: str, data: dict[str, Any]
     ) -> FiefUserInfo:
         """
         Update user information with the Fief API using a valid access token.
@@ -917,18 +1165,6 @@ class FiefAsync(BaseFief):
         :param access_token: A valid access token.
         :param data: A dictionary containing the data to update.
 
-        **Example: Update email address**
-
-        ```py
-        userinfo = await fief.update_profile("ACCESS_TOKEN", { "email": "anne@nantes.city" })
-        ```
-
-        **Example: Update password**
-
-        ```py
-        userinfo = await fief.update_profile("ACCESS_TOKEN", { "password": "hermine1" })
-        ```
-
         **Example: Update user field**
 
         To update [user field](https://docs.fief.dev/getting-started/user-fields/) values, you need to nest them into a `fields` dictionary, indexed by their slug.
@@ -948,7 +1184,102 @@ class FiefAsync(BaseFief):
             )
             response = await client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
+
+            return response.json()
+
+    async def change_password(
+        self, access_token: str, new_password: str
+    ) -> FiefUserInfo:
+        """
+        Change the user password with the Fief API using a valid access token.
+
+        **An access token with an ACR of at least level 1 is required.**
+
+        :param access_token: A valid access token.
+        :param new_password: The new password.
+
+        **Example**
+
+        ```py
+        userinfo = await fief.change_password("ACCESS_TOKEN", "herminetincture")
+        ```
+        """
+        change_password_profile_endpoint = f"{self.base_url}/api/password"
+
+        async with self._get_httpx_client() as client:
+            request = self._get_change_password_request(
+                client,
+                endpoint=change_password_profile_endpoint,
+                access_token=access_token,
+                new_password=new_password,
+            )
+            response = await client.send(request)
+
+            self._handle_request_error(response)
+
+            return response.json()
+
+    async def email_change(self, access_token: str, email: str) -> FiefUserInfo:
+        """
+        Request an email change with the Fief API using a valid access token.
+
+        The user will receive a verification code on this new email address.
+        It shall be used with the method `email_verify` to complete the modification.
+
+        **An access token with an ACR of at least level 1 is required.**
+
+        :param access_token: A valid access token.
+        :param email: The new email address.
+
+        **Example**
+
+        ```py
+        userinfo = await fief.email_change("ACCESS_TOKEN", "anne@nantes.city")
+        ```
+        """
+        email_change_endpoint = f"{self.base_url}/api/email/change"
+
+        async with self._get_httpx_client() as client:
+            request = self._get_email_change_request(
+                client,
+                endpoint=email_change_endpoint,
+                access_token=access_token,
+                email=email,
+            )
+            response = await client.send(request)
+
+            self._handle_request_error(response)
+
+            return response.json()
+
+    async def email_verify(self, access_token: str, code: str) -> FiefUserInfo:
+        """
+        Verify the user email with the Fief API using a valid access token and verification code.
+
+        **An access token with an ACR of at least level 1 is required.**
+
+        :param access_token: A valid access token.
+        :param code: The verification code received by email.
+
+        **Example**
+
+        ```py
+        userinfo = fief.email_verify("ACCESS_TOKEN", "ABCDE")
+        ```
+        """
+        email_verify_endpoint = f"{self.base_url}/api/email/verify"
+
+        async with self._get_httpx_client() as client:
+            request = self._get_email_verify_request(
+                client,
+                endpoint=email_verify_endpoint,
+                access_token=access_token,
+                code=code,
+            )
+            response = await client.send(request)
+
+            self._handle_request_error(response)
 
             return response.json()
 
@@ -980,7 +1311,7 @@ class FiefAsync(BaseFief):
         ) as client:
             yield client
 
-    async def _get_openid_configuration(self) -> Dict[str, Any]:
+    async def _get_openid_configuration(self) -> dict[str, Any]:
         if self._openid_configuration is not None:
             return self._openid_configuration
 
@@ -1019,6 +1350,6 @@ class FiefAsync(BaseFief):
             )
             response = await client.send(request)
 
-            response.raise_for_status()
+            self._handle_request_error(response)
 
             return response.json()
