@@ -1,24 +1,17 @@
 """FastAPI integration."""
-import sys
+
 import uuid
+from collections.abc import AsyncGenerator, Coroutine, Generator
 from inspect import Parameter, Signature, isawaitable
 from typing import (
-    AsyncGenerator,
     Callable,
-    Coroutine,
-    Generator,
-    List,
     Optional,
+    Protocol,
     TypeVar,
     Union,
     cast,
 )
 
-if sys.version_info < (3, 8):
-    from typing_extensions import Protocol  # pragma: no cover
-else:
-    from typing import Protocol  # pragma: no cover
-
 from fastapi import Depends, HTTPException, Request, Response, status
 from fastapi.security.base import SecurityBase
 from fastapi.security.http import HTTPAuthorizationCredentials
@@ -26,11 +19,13 @@ from makefun import with_signature
 
 from fief_client import (
     Fief,
+    FiefAccessTokenACRTooLow,
     FiefAccessTokenExpired,
     FiefAccessTokenInfo,
     FiefAccessTokenInvalid,
     FiefAccessTokenMissingPermission,
     FiefAccessTokenMissingScope,
+    FiefACR,
     FiefAsync,
     FiefUserInfo,
 )
@@ -109,7 +104,7 @@ class FiefAuth:
         client: FiefClientClass,
         scheme: SecurityBase,
         *,
-        get_userinfo_cache: Optional[DependencyCallable[UserInfoCacheProtocol]] = None
+        get_userinfo_cache: Optional[DependencyCallable[UserInfoCacheProtocol]] = None,
     ) -> None:
         """
         :param client: Instance of a Fief client.
@@ -128,8 +123,9 @@ class FiefAuth:
     def authenticated(
         self,
         optional: bool = False,
-        scope: Optional[List[str]] = None,
-        permissions: Optional[List[str]] = None,
+        scope: Optional[list[str]] = None,
+        acr: Optional[FiefACR] = None,
+        permissions: Optional[list[str]] = None,
     ):
         """
         Return a FastAPI dependency to check if a request is authenticated.
@@ -140,6 +136,9 @@ class FiefAuth:
         an unauthorized response will be raised.
         :param scope: Optional list of scopes required.
         If the access token lacks one of the required scope, a forbidden response will be raised.
+        :param acr: Optional minimum ACR level required.
+        If the access token doesn't meet the minimum level, a forbidden response will be raised.
+        Read more: https://docs.fief.dev/going-further/acr/
         :param permissions: Optional list of permissions required.
         If the access token lacks one of the required permission, a forbidden response will be raised.
 
@@ -169,15 +168,24 @@ class FiefAuth:
 
             try:
                 result = self.client.validate_access_token(
-                    token, required_scope=scope, required_permissions=permissions
+                    token,
+                    required_scope=scope,
+                    required_acr=acr,
+                    required_permissions=permissions,
                 )
                 if isawaitable(result):
                     info = await result
                 else:
                     info = result
             except (FiefAccessTokenInvalid, FiefAccessTokenExpired):
+                if optional:
+                    return None
                 return await self.get_unauthorized_response(request, response)
-            except (FiefAccessTokenMissingScope, FiefAccessTokenMissingPermission):
+            except (
+                FiefAccessTokenMissingScope,
+                FiefAccessTokenACRTooLow,
+                FiefAccessTokenMissingPermission,
+            ):
                 return await self.get_forbidden_response(request, response)
 
             return info
@@ -187,8 +195,9 @@ class FiefAuth:
     def current_user(
         self,
         optional: bool = False,
-        scope: Optional[List[str]] = None,
-        permissions: Optional[List[str]] = None,
+        scope: Optional[list[str]] = None,
+        acr: Optional[FiefACR] = None,
+        permissions: Optional[list[str]] = None,
         refresh: bool = False,
     ):
         """
@@ -202,6 +211,9 @@ class FiefAuth:
         an unauthorized response will be raised.
         :param scope: Optional list of scopes required.
         If the access token lacks one of the required scope, a forbidden response will be raised.
+        :param acr: Optional minimum ACR level required.
+        If the access token doesn't meet the minimum level, a forbidden response will be raised.
+        Read more: https://docs.fief.dev/going-further/acr/
         :param permissions: Optional list of permissions required.
         If the access token lacks one of the required permission, a forbidden response will be raised.
         :param refresh: If `True`, the user information will be refreshed from the Fief API.
@@ -218,7 +230,7 @@ class FiefAuth:
         ```
         """
         signature = self._get_current_user_call_signature(
-            self.authenticated(optional, scope, permissions)
+            self.authenticated(optional, scope, acr, permissions)
         )
 
         @with_signature(signature)
@@ -281,7 +293,7 @@ class FiefAuth:
         with a dynamic security scheme dependency at runtime.
         This way, it's detected by the OpenAPI generator.
         """
-        parameters: List[Parameter] = [
+        parameters: list[Parameter] = [
             Parameter(
                 name="request",
                 kind=Parameter.POSITIONAL_OR_KEYWORD,
@@ -309,7 +321,7 @@ class FiefAuth:
         with a dynamic security scheme dependency at runtime.
         This way, it's detected by the OpenAPI generator.
         """
-        parameters: List[Parameter] = [
+        parameters: list[Parameter] = [
             Parameter(
                 name="access_token_info",
                 kind=Parameter.POSITIONAL_OR_KEYWORD,
