@@ -1,7 +1,9 @@
+import contextlib
 import uuid
+from collections.abc import Generator
 from datetime import datetime, timezone
 from os import path
-from typing import Callable, Generator, List
+from typing import Callable, Protocol
 
 import pytest
 import pytest_asyncio
@@ -9,10 +11,12 @@ import respx
 from httpx import Response
 from jwcrypto import jwk, jwt
 
+from fief_client import FiefACR
+
 
 @pytest.fixture(scope="session")
 def keys() -> jwk.JWKSet:
-    with open(path.join(path.dirname(__file__), "jwks.json"), "r") as jwks_file:
+    with open(path.join(path.dirname(__file__), "jwks.json")) as jwks_file:
         return jwk.JWKSet.from_json(jwks_file.read())
 
 
@@ -67,10 +71,15 @@ def generate_token(signature_key: jwk.JWK, encryption_key: jwk.JWK, user_id: str
 @pytest.fixture(scope="session")
 def generate_access_token(generate_token: Callable[..., str]):
     def _generate_access_token(
-        encrypt: bool, *, scope: str = "", permissions: List[str] = [], **kwargs
+        encrypt: bool,
+        *,
+        scope: str = "",
+        permissions: list[str] = [],
+        acr: FiefACR = FiefACR.LEVEL_ZERO,
+        **kwargs,
     ) -> str:
         return generate_token(
-            encrypt=encrypt, scope=scope, permissions=permissions, **kwargs
+            encrypt=encrypt, scope=scope, permissions=permissions, acr=acr, **kwargs
         )
 
     return _generate_access_token
@@ -91,27 +100,49 @@ def encrypted_id_token(generate_token: Callable[..., str]) -> str:
     return generate_token(encrypt=True)
 
 
+class GetAPIRequestsMock(Protocol):
+    def __call__(
+        self, *, hostname: str = "https://bretagne.fief.dev", path_prefix: str = ""
+    ) -> contextlib.AbstractContextManager[respx.MockRouter]: ...
+
+
+@pytest_asyncio.fixture(scope="module")
+def get_api_requests_mock(signature_key: jwk.JWK) -> GetAPIRequestsMock:
+    @contextlib.contextmanager
+    def _get_api_requests_mock(
+        *, hostname: str = "https://bretagne.fief.dev", path_prefix: str = ""
+    ) -> Generator[respx.MockRouter, None, None]:
+        with respx.mock(assert_all_mocked=True, assert_all_called=False) as respx_mock:
+            openid_configuration_route = respx_mock.get(
+                f"{path_prefix}/.well-known/openid-configuration"
+            )
+            openid_configuration_route.return_value = Response(
+                200,
+                json={
+                    "issuer": f"{hostname}{path_prefix}",
+                    "authorization_endpoint": f"{hostname}{path_prefix}/authorize",
+                    "token_endpoint": f"{hostname}{path_prefix}/token",
+                    "userinfo_endpoint": f"{hostname}{path_prefix}/userinfo",
+                    "jwks_uri": f"{hostname}{path_prefix}/.well-known/jwks.json",
+                },
+            )
+
+            jwks_route = respx_mock.get(f"{path_prefix}/.well-known/jwks.json")
+            jwks_route.return_value = Response(
+                200,
+                json={"keys": [signature_key.export(private_key=False, as_dict=True)]},
+            )
+
+            print("YIELD")
+            yield respx_mock
+        print("RESET")
+
+    return _get_api_requests_mock
+
+
 @pytest_asyncio.fixture(scope="module", autouse=True)
 def mock_api_requests(
-    signature_key: jwk.JWK,
+    get_api_requests_mock: GetAPIRequestsMock,
 ) -> Generator[respx.MockRouter, None, None]:
-    HOSTNAME = "https://bretagne.fief.dev"
-
-    with respx.mock(assert_all_mocked=True, assert_all_called=False) as respx_mock:
-        openid_configuration_route = respx_mock.get("/.well-known/openid-configuration")
-        openid_configuration_route.return_value = Response(
-            200,
-            json={
-                "authorization_endpoint": f"{HOSTNAME}/authorize",
-                "token_endpoint": f"{HOSTNAME}/token",
-                "userinfo_endpoint": f"{HOSTNAME}/userinfo",
-                "jwks_uri": f"{HOSTNAME}/.well-known/jwks.json",
-            },
-        )
-
-        jwks_route = respx_mock.get("/.well-known/jwks.json")
-        jwks_route.return_value = Response(
-            200, json={"keys": [signature_key.export(private_key=False, as_dict=True)]}
-        )
-
+    with get_api_requests_mock() as respx_mock:
         yield respx_mock
