@@ -1,7 +1,8 @@
 import contextlib
 import json
 import uuid
-from typing import Dict, List, Mapping, Optional
+from collections.abc import Mapping
+from typing import Optional
 
 import httpx
 import pytest
@@ -12,15 +13,19 @@ from pytest_mock import MockerFixture
 
 from fief_client.client import (
     Fief,
+    FiefAccessTokenACRTooLow,
     FiefAccessTokenExpired,
     FiefAccessTokenInvalid,
     FiefAccessTokenMissingPermission,
     FiefAccessTokenMissingScope,
+    FiefACR,
     FiefAsync,
     FiefIdTokenInvalid,
+    FiefRequestError,
     FiefTokenResponse,
 )
 from fief_client.crypto import get_validation_hash
+from tests.conftest import GetAPIRequestsMock
 
 
 @pytest.fixture(scope="module")
@@ -28,6 +33,11 @@ def fief_client() -> Fief:
     return Fief("https://bretagne.fief.dev", "CLIENT_ID", "CLIENT_SECRET")
 
 
+@pytest.fixture(scope="module")
+def fief_client_tenant() -> Fief:
+    return Fief("https://bretagne.fief.dev/secondary", "CLIENT_ID", "CLIENT_SECRET")
+
+
 @pytest.fixture(scope="module")
 def fief_client_encryption_key(encryption_key: jwk.JWK) -> Fief:
     return Fief(
@@ -57,6 +67,14 @@ def test_serializable_fief_token_response():
     )
 
 
+def test_fief_acr():
+    assert FiefACR.LEVEL_ZERO < FiefACR.LEVEL_ONE
+    assert FiefACR.LEVEL_ZERO <= FiefACR.LEVEL_ONE
+    assert FiefACR.LEVEL_ONE > FiefACR.LEVEL_ZERO
+    assert FiefACR.LEVEL_ONE >= FiefACR.LEVEL_ZERO
+    assert FiefACR.LEVEL_ZERO == FiefACR.LEVEL_ZERO
+
+
 class TestCustomVerifyCertParameters:
     def test_sync(self, mocker: MockerFixture):
         client_mock = mocker.patch.object(httpx, "Client")
@@ -133,7 +151,7 @@ class TestAuthURL:
     def test_authorization_url(
         self,
         state: Optional[str],
-        scope: Optional[List[str]],
+        scope: Optional[list[str]],
         code_challenge: Optional[str],
         code_challenge_method: Optional[str],
         lang: Optional[str],
@@ -194,7 +212,7 @@ class TestAuthURL:
     async def test_authorization_url_async(
         self,
         state: Optional[str],
-        scope: Optional[List[str]],
+        scope: Optional[list[str]],
         code_challenge: Optional[str],
         code_challenge_method: Optional[str],
         lang: Optional[str],
@@ -224,8 +242,36 @@ class TestAuthURL:
 
         assert request.url.host == request.headers["Host"]
 
+    def test_authorization_url_tenant(
+        self, fief_client_tenant: Fief, get_api_requests_mock: GetAPIRequestsMock
+    ):
+        with get_api_requests_mock(path_prefix="/secondary"):
+            authorize_url = fief_client_tenant.auth_url(
+                "https://www.bretagne.duchy/callback"
+            )
+
+        assert (
+            authorize_url
+            == "https://bretagne.fief.dev/secondary/authorize?response_type=code&client_id=CLIENT_ID&redirect_uri=https%3A%2F%2Fwww.bretagne.duchy%2Fcallback"
+        )
+
 
 class TestAuthCallback:
+    def test_error_response(
+        self, fief_client: Fief, mock_api_requests: respx.MockRouter
+    ):
+        token_route = mock_api_requests.post("/token")
+        token_route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            fief_client.auth_callback(
+                "CODE",
+                "https://www.bretagne.duchy/callback",
+                code_verifier="CODE_VERIFIER",
+            )
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     def test_valid_response(
         self,
         fief_client: Fief,
@@ -250,7 +296,10 @@ class TestAuthCallback:
 
         token_route_call = token_route.calls.last
         assert token_route_call is not None
-        assert "Authorization" in token_route_call.request.headers
+
+        request_data = token_route_call.request.content.decode("utf-8")
+        assert "client_id" in request_data
+        assert "client_secret" in request_data
 
         assert token_response["access_token"] == access_token
         assert token_response["id_token"] == signed_id_token
@@ -258,6 +307,22 @@ class TestAuthCallback:
         assert isinstance(userinfo, dict)
         assert userinfo["sub"] == user_id
 
+    @pytest.mark.asyncio
+    async def test_error_response_async(
+        self, fief_async_client: FiefAsync, mock_api_requests: respx.MockRouter
+    ):
+        token_route = mock_api_requests.post("/token")
+        token_route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            await fief_async_client.auth_callback(
+                "CODE",
+                "https://www.bretagne.duchy/callback",
+                code_verifier="CODE_VERIFIER",
+            )
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     @pytest.mark.asyncio
     async def test_valid_response_async(
         self,
@@ -283,7 +348,10 @@ class TestAuthCallback:
 
         token_route_call = token_route.calls.last
         assert token_route_call is not None
-        assert "Authorization" in token_route_call.request.headers
+
+        request_data = token_route_call.request.content.decode("utf-8")
+        assert "client_id" in request_data
+        assert "client_secret" in request_data
 
         assert token_response["access_token"] == access_token
         assert token_response["id_token"] == signed_id_token
@@ -291,8 +359,89 @@ class TestAuthCallback:
         assert isinstance(userinfo, dict)
         assert userinfo["sub"] == user_id
 
+    def test_valid_response_tenant(
+        self,
+        fief_client_tenant: Fief,
+        get_api_requests_mock: GetAPIRequestsMock,
+        access_token: str,
+        signed_id_token: str,
+        user_id: str,
+    ):
+        with get_api_requests_mock(path_prefix="/secondary") as mock_api_requests:
+            token_route = mock_api_requests.post("/secondary/token")
+            token_route.return_value = Response(
+                200,
+                json={
+                    "access_token": access_token,
+                    "id_token": signed_id_token,
+                    "token_type": "bearer",
+                },
+            )
+
+            token_response, userinfo = fief_client_tenant.auth_callback(
+                "CODE",
+                "https://www.bretagne.duchy/callback",
+                code_verifier="CODE_VERIFIER",
+            )
+
+            token_route_call = token_route.calls.last
+            assert token_route_call is not None
+
+            request_data = token_route_call.request.content.decode("utf-8")
+            assert "client_id" in request_data
+            assert "client_secret" in request_data
+
+            assert token_response["access_token"] == access_token
+            assert token_response["id_token"] == signed_id_token
+
+            assert isinstance(userinfo, dict)
+            assert userinfo["sub"] == user_id
+
+    def test_no_code_verifier(
+        self,
+        fief_client: Fief,
+        mock_api_requests: respx.MockRouter,
+        access_token: str,
+        signed_id_token: str,
+        user_id: str,
+    ):
+        token_route = mock_api_requests.post("/token")
+        token_route.return_value = Response(
+            200,
+            json={
+                "access_token": access_token,
+                "id_token": signed_id_token,
+                "token_type": "bearer",
+            },
+        )
+
+        token_response, userinfo = fief_client.auth_callback(
+            "CODE", "https://www.bretagne.duchy/callback"
+        )
+
+        token_route_call = token_route.calls.last
+        assert token_route_call is not None
+
+        request_data = token_route_call.request.content.decode("utf-8")
+        assert "client_id" in request_data
+        assert "client_secret" in request_data
+        assert "code_verifier" not in request_data
+
 
 class TestAuthRefreshToken:
+    def test_error_response(
+        self, fief_client: Fief, mock_api_requests: respx.MockRouter
+    ):
+        token_route = mock_api_requests.post("/token")
+        token_route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            fief_client.auth_refresh_token(
+                "REFRESH_TOKEN", scope=["openid", "offline_access"]
+            )
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     def test_valid_response(
         self,
         fief_client: Fief,
@@ -317,7 +466,10 @@ class TestAuthRefreshToken:
 
         token_route_call = token_route.calls.last
         assert token_route_call is not None
-        assert "Authorization" in token_route_call.request.headers
+
+        request_data = token_route_call.request.content.decode("utf-8")
+        assert "client_id" in request_data
+        assert "client_secret" in request_data
 
         assert token_response["access_token"] == access_token
         assert token_response["id_token"] == signed_id_token
@@ -325,6 +477,20 @@ class TestAuthRefreshToken:
         assert isinstance(userinfo, dict)
         assert userinfo["sub"] == user_id
 
+    @pytest.mark.asyncio
+    async def test_error_response_async(
+        self, fief_async_client: FiefAsync, mock_api_requests: respx.MockRouter
+    ):
+        token_route = mock_api_requests.post("/token")
+        token_route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            await fief_async_client.auth_refresh_token(
+                "REFRESH_TOKEN", scope=["openid", "offline_access"]
+            )
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     @pytest.mark.asyncio
     async def test_valid_response_async(
         self,
@@ -350,7 +516,10 @@ class TestAuthRefreshToken:
 
         token_route_call = token_route.calls.last
         assert token_route_call is not None
-        assert "Authorization" in token_route_call.request.headers
+
+        request_data = token_route_call.request.content.decode("utf-8")
+        assert "client_id" in request_data
+        assert "client_secret" in request_data
 
         assert token_response["access_token"] == access_token
         assert token_response["id_token"] == signed_id_token
@@ -360,6 +529,10 @@ class TestAuthRefreshToken:
 
 
 class TestValidateAccessToken:
+    def test_invalid_token(self, fief_client: Fief):
+        with pytest.raises(FiefAccessTokenInvalid):
+            fief_client.validate_access_token("INVALID_TOKEN")
+
     def test_invalid_signature(self, fief_client: Fief):
         with pytest.raises(FiefAccessTokenInvalid):
             fief_client.validate_access_token(
@@ -372,6 +545,11 @@ class TestValidateAccessToken:
                 "eyJhbGciOiJSUzI1NiJ9.e30.RmKxjgPljzJL_-Yp9oBJIvNejvES_pnTeZBDvptYcdWm4Ze9D6FlM8RFJ5-ZJ3O-HXlWylVXiGAE_wdSGXehSaENUN3Mj91j5OfiXGrtBGSiEiCtC9HYKCi6xf6xmcEPoTbtBVi38a9OARoJlpTJ5T4BbmqIUR8R06sqo3zTkwk48wPmYtk_OPgMv4c8tNyHF17dRe1JM_ix-m7V1Nv_2DHLMRgMXdsWkl0RCcAFQwqCTXU4UxWSoXp6CB0-Ybkq-P5KyXIXy0b15qG8jfgCrFHqFhN3hpyvL4Zza_EkXJaCkB5v-oztlHS6gTGb3QgFqppW3JM6TJnDKslGRPDsjg"
             )
 
+    def test_invalid_acr_claim(self, fief_client: Fief, generate_access_token):
+        access_token = generate_access_token(encrypt=False, acr="INVALID_ACR")
+        with pytest.raises(FiefAccessTokenInvalid):
+            fief_client.validate_access_token(access_token)
+
     def test_expired(self, fief_client: Fief, generate_access_token):
         access_token = generate_access_token(encrypt=False, exp=0)
         with pytest.raises(FiefAccessTokenExpired):
@@ -394,6 +572,27 @@ class TestValidateAccessToken:
         assert info == {
             "id": uuid.UUID(user_id),
             "scope": ["openid", "offline_access"],
+            "acr": FiefACR.LEVEL_ZERO,
+            "permissions": [],
+            "access_token": access_token,
+        }
+
+    def test_invalid_acr(self, fief_client: Fief, generate_access_token):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ZERO)
+        with pytest.raises(FiefAccessTokenACRTooLow):
+            fief_client.validate_access_token(
+                access_token, required_acr=FiefACR.LEVEL_ONE
+            )
+
+    def test_valid_acr(self, fief_client: Fief, generate_access_token, user_id: str):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ONE)
+        info = fief_client.validate_access_token(
+            access_token, required_acr=FiefACR.LEVEL_ONE
+        )
+        assert info == {
+            "id": uuid.UUID(user_id),
+            "scope": [],
+            "acr": FiefACR.LEVEL_ONE,
             "permissions": [],
             "access_token": access_token,
         }
@@ -419,10 +618,16 @@ class TestValidateAccessToken:
         assert info == {
             "id": uuid.UUID(user_id),
             "scope": [],
+            "acr": FiefACR.LEVEL_ZERO,
             "permissions": ["castles:read", "castles:create"],
             "access_token": access_token,
         }
 
+    @pytest.mark.asyncio
+    async def test_async_invalid_token(self, fief_async_client: FiefAsync):
+        with pytest.raises(FiefAccessTokenInvalid):
+            await fief_async_client.validate_access_token("INVALID_TOKEN")
+
     @pytest.mark.asyncio
     async def test_async_invalid_signature(self, fief_async_client: FiefAsync):
         with pytest.raises(FiefAccessTokenInvalid):
@@ -470,6 +675,33 @@ class TestValidateAccessToken:
         assert info == {
             "id": uuid.UUID(user_id),
             "scope": ["openid", "offline_access"],
+            "acr": FiefACR.LEVEL_ZERO,
+            "permissions": [],
+            "access_token": access_token,
+        }
+
+    @pytest.mark.asyncio
+    async def test_async_invalid_acr(
+        self, fief_async_client: FiefAsync, generate_access_token
+    ):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ZERO)
+        with pytest.raises(FiefAccessTokenACRTooLow):
+            await fief_async_client.validate_access_token(
+                access_token, required_acr=FiefACR.LEVEL_ONE
+            )
+
+    @pytest.mark.asyncio
+    async def test_async_valid_acr(
+        self, fief_async_client: FiefAsync, generate_access_token, user_id: str
+    ):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ONE)
+        info = await fief_async_client.validate_access_token(
+            access_token, required_acr=FiefACR.LEVEL_ONE
+        )
+        assert info == {
+            "id": uuid.UUID(user_id),
+            "scope": [],
+            "acr": FiefACR.LEVEL_ONE,
             "permissions": [],
             "access_token": access_token,
         }
@@ -499,12 +731,24 @@ class TestValidateAccessToken:
         assert info == {
             "id": uuid.UUID(user_id),
             "scope": [],
+            "acr": FiefACR.LEVEL_ZERO,
             "permissions": ["castles:read", "castles:create"],
             "access_token": access_token,
         }
 
 
 class TestUserinfo:
+    def test_error_response(
+        self, fief_client: Fief, mock_api_requests: respx.MockRouter
+    ):
+        token_route = mock_api_requests.get("/userinfo")
+        token_route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            fief_client.userinfo("ACCESS_TOKEN")
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     def test_valid_response(
         self, fief_client: Fief, mock_api_requests: respx.MockRouter, user_id: str
     ):
@@ -515,6 +759,18 @@ class TestUserinfo:
         userinfo = fief_client.userinfo("ACCESS_TOKEN")
         assert userinfo == {"sub": user_id}
 
+    @pytest.mark.asyncio
+    async def test_error_response_async(
+        self, fief_async_client: FiefAsync, mock_api_requests: respx.MockRouter
+    ):
+        token_route = mock_api_requests.get("/userinfo")
+        token_route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            await fief_async_client.userinfo("ACCESS_TOKEN")
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     @pytest.mark.asyncio
     async def test_valid_response_async(
         self,
@@ -530,33 +786,88 @@ class TestUserinfo:
         assert userinfo == {"sub": user_id}
 
 
-class TestUpdateProfile:
-    def test_valid_response(
-        self, fief_client: Fief, mock_api_requests: respx.MockRouter, user_id: str
+@pytest.mark.parametrize(
+    "endpoint,method_name,args",
+    [
+        (
+            "/api/profile",
+            "update_profile",
+            ("ACCESS_TOKEN", {"fields": {"first_name": "Anne"}}),
+        ),
+        ("/api/password", "change_password", ("ACCESS_TOKEN", "herminetincture")),
+        ("/api/email/change", "email_change", ("ACCESS_TOKEN", "anne@nantes.city")),
+        ("/api/email/verify", "email_verify", ("ACCESS_TOKEN", "ABCDE")),
+    ],
+)
+class TestUpdateUserMethods:
+    def test_error_response(
+        self,
+        endpoint: str,
+        method_name: str,
+        args: tuple,
+        fief_client: Fief,
+        mock_api_requests: respx.MockRouter,
     ):
-        mock_api_requests.patch("/api/profile").return_value = Response(
+        route = mock_api_requests.route(path=endpoint)
+        route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            method = getattr(fief_client, method_name)
+            method(*args)
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
+    def test_valid_response(
+        self,
+        endpoint: str,
+        method_name: str,
+        args: tuple,
+        fief_client: Fief,
+        mock_api_requests: respx.MockRouter,
+        user_id: str,
+    ):
+        mock_api_requests.route(path=endpoint).return_value = Response(
             200, json={"sub": user_id}
         )
 
-        userinfo = fief_client.update_profile(
-            "ACCESS_TOKEN", {"email": "anne@bretagne.duchy"}
-        )
+        method = getattr(fief_client, method_name)
+        userinfo = method(*args)
         assert userinfo == {"sub": user_id}
 
+    @pytest.mark.asyncio
+    async def test_error_response_async(
+        self,
+        endpoint: str,
+        method_name: str,
+        args: tuple,
+        fief_async_client: FiefAsync,
+        mock_api_requests: respx.MockRouter,
+    ):
+        route = mock_api_requests.route(path=endpoint)
+        route.return_value = Response(400, json={"detail": "error"})
+
+        with pytest.raises(FiefRequestError) as excinfo:
+            method = getattr(fief_async_client, method_name)
+            await method(*args)
+        assert excinfo.value.status_code == 400
+        assert excinfo.value.detail == '{"detail": "error"}'
+
     @pytest.mark.asyncio
     async def test_valid_response_async(
         self,
+        endpoint: str,
+        method_name: str,
+        args: tuple,
         fief_async_client: FiefAsync,
         mock_api_requests: respx.MockRouter,
         user_id: str,
     ):
-        mock_api_requests.patch("/api/profile").return_value = Response(
+        mock_api_requests.route(path=endpoint).return_value = Response(
             200, json={"sub": user_id}
         )
 
-        userinfo = await fief_async_client.update_profile(
-            "ACCESS_TOKEN", {"email": "anne@bretagne.duchy"}
-        )
+        method = getattr(fief_async_client, method_name)
+        userinfo = await method(*args)
         assert userinfo == {"sub": user_id}
 
 
@@ -652,7 +963,7 @@ class TestDecodeIdToken:
     )
     def test_signed_at_hash_c_hash_invalid(
         self,
-        claims: Dict[str, str],
+        claims: dict[str, str],
         fief_client: Fief,
         signature_key: jwk.JWK,
         generate_token,
