@@ -1,5 +1,6 @@
 import uuid
-from typing import Generator, Optional
+from collections.abc import Generator
+from typing import Optional
 
 import pytest
 import respx
@@ -7,7 +8,7 @@ from flask import Flask, g, session
 from flask.testing import FlaskClient
 from httpx import Response
 
-from fief_client.client import Fief, FiefUserInfo
+from fief_client.client import Fief, FiefACR, FiefUserInfo
 from fief_client.integrations.flask import (
     FiefAuth,
     FiefAuthForbidden,
@@ -63,6 +64,11 @@ def flask_app(fief_client: Fief) -> Generator[Flask, None, None]:
     def get_authenticated_scope():
         return g.access_token_info
 
+    @app.get("/authenticated-acr")
+    @auth.authenticated(acr=FiefACR.LEVEL_ONE)
+    def get_authenticated_acr():
+        return g.access_token_info
+
     @app.get("/authenticated-permission")
     @auth.authenticated(permissions=["castles:create"])
     def get_authenticated_permission():
@@ -88,6 +94,11 @@ def flask_app(fief_client: Fief) -> Generator[Flask, None, None]:
     def get_current_user_scope():
         return g.user
 
+    @app.get("/current-user-acr")
+    @auth.current_user(acr=FiefACR.LEVEL_ONE)
+    def get_current_user_acr():
+        return g.user
+
     @app.get("/current-user-permission")
     @auth.current_user(permissions=["castles:create"])
     def get_current_user_permission():
@@ -112,6 +123,13 @@ class TestAuthenticated:
 
         assert response.status_code == 401
 
+    def test_invalid_token(self, test_client: FlaskClient):
+        response = test_client.get(
+            "/authenticated", headers={"Authorization": "Bearer INVALID_TOKEN"}
+        )
+
+        assert response.status_code == 401
+
     def test_expired_token(self, test_client: FlaskClient, generate_access_token):
         access_token = generate_access_token(encrypt=False, exp=0)
 
@@ -136,6 +154,7 @@ class TestAuthenticated:
         assert json == {
             "id": user_id,
             "scope": ["openid"],
+            "acr": FiefACR.LEVEL_ZERO,
             "permissions": [],
             "access_token": access_token,
         }
@@ -147,18 +166,26 @@ class TestAuthenticated:
         assert response.status_code == 200
         assert response.json == {}
 
-        access_token = generate_access_token(encrypt=False, scope="openid")
+        expired_access_token = generate_access_token(
+            encrypt=False, scope="openid", exp=0
+        )
+        response = test_client.get(
+            "/authenticated-optional",
+            headers={"Authorization": f"Bearer {expired_access_token}"},
+        )
+        assert response.status_code == 200
+        assert response.json == {}
 
+        access_token = generate_access_token(encrypt=False, scope="openid")
         response = test_client.get(
             "/authenticated-optional",
             headers={"Authorization": f"Bearer {access_token}"},
         )
-
         assert response.status_code == 200
-
         assert response.json == {
             "id": user_id,
             "scope": ["openid"],
+            "acr": FiefACR.LEVEL_ZERO,
             "permissions": [],
             "access_token": access_token,
         }
@@ -189,6 +216,36 @@ class TestAuthenticated:
         assert json == {
             "id": user_id,
             "scope": ["openid", "required_scope"],
+            "acr": FiefACR.LEVEL_ZERO,
+            "permissions": [],
+            "access_token": access_token,
+        }
+
+    def test_invalid_acr(self, test_client: FlaskClient, generate_access_token):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ZERO)
+
+        response = test_client.get(
+            "/authenticated-acr", headers={"Authorization": f"Bearer {access_token}"}
+        )
+
+        assert response.status_code == 403
+
+    def test_valid_acr(
+        self, test_client: FlaskClient, generate_access_token, user_id: str
+    ):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ONE)
+
+        response = test_client.get(
+            "/authenticated-acr", headers={"Authorization": f"Bearer {access_token}"}
+        )
+
+        assert response.status_code == 200
+
+        json = response.json
+        assert json == {
+            "id": user_id,
+            "scope": [],
+            "acr": FiefACR.LEVEL_ONE,
             "permissions": [],
             "access_token": access_token,
         }
@@ -223,6 +280,7 @@ class TestAuthenticated:
         assert json == {
             "id": user_id,
             "scope": [],
+            "acr": FiefACR.LEVEL_ZERO,
             "permissions": ["castles:read", "castles:create"],
             "access_token": access_token,
         }
@@ -299,15 +357,22 @@ class TestCurrentUser:
         assert response.status_code == 200
         assert response.json == {}
 
-        access_token = generate_access_token(encrypt=False, scope="openid")
+        expired_access_token = generate_access_token(
+            encrypt=False, scope="openid", exp=0
+        )
+        response = test_client.get(
+            "/current-user-optional",
+            headers={"Authorization": f"Bearer {expired_access_token}"},
+        )
+        assert response.status_code == 200
+        assert response.json == {}
 
+        access_token = generate_access_token(encrypt=False, scope="openid")
         response = test_client.get(
             "/current-user-optional",
             headers={"Authorization": f"Bearer {access_token}"},
         )
-
         assert response.status_code == 200
-
         assert response.json == {"sub": user_id}
 
     def test_missing_scope(self, test_client: FlaskClient, generate_access_token):
@@ -343,6 +408,37 @@ class TestCurrentUser:
         json = response.json
         assert json == {"sub": user_id}
 
+    def test_missing_acr(self, test_client: FlaskClient, generate_access_token):
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ZERO)
+
+        response = test_client.get(
+            "/current-user-acr", headers={"Authorization": f"Bearer {access_token}"}
+        )
+
+        assert response.status_code == 403
+
+    def test_valid_acr(
+        self,
+        test_client: FlaskClient,
+        generate_access_token,
+        mock_api_requests: respx.MockRouter,
+        user_id: str,
+    ):
+        mock_api_requests.get("/userinfo").return_value = Response(
+            200, json={"sub": user_id}
+        )
+
+        access_token = generate_access_token(encrypt=False, acr=FiefACR.LEVEL_ONE)
+
+        response = test_client.get(
+            "/current-user-acr", headers={"Authorization": f"Bearer {access_token}"}
+        )
+
+        assert response.status_code == 200
+
+        json = response.json
+        assert json == {"sub": user_id}
+
     def test_missing_permission(self, test_client: FlaskClient, generate_access_token):
         access_token = generate_access_token(
             encrypt=False, permissions=["castles:read"]
